/*
 *  app_flite -- A Flite TTS engine interface  
 *
 *  Copyright (C) 2008, Darren Sessions
 *  Copyright (C) 2007, Francois Aucamp
 *
 *  Darren Sessions <dmsessions@gmail.com>
 *
 *
 *  This program is free software, distributed under the terms of
 *  the GNU General Public License Version 2. Read the LICENSE 
 *  file for details.
 *
 */

/*! \file
 *
 * \brief Flite TTS engine interface
 *
 * \author Darren Sessions <dmsessions@gmail.com>
 *
 * \ingroup applications
 */

/*** MODULEINFO
	<defaultenabled>no</defaultenabled>
 ***/
#define AST_MODULE "flite"

#if ASTERISK_VERSION >= 6
#include "asterisk.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "asterisk/astobj.h"
#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/translate.h>
#include <asterisk/utils.h>
#include <asterisk/md5.h>

#include <flite/flite.h>

#define FLITE_CONFIG "flite.conf"

cst_voice *register_cmu_us_kal(void);

#if ASTERISK_VERSION < 4
static char *tdesc = "Flite Text-to-Speech Interface";
#endif

static char *app = "Flite";

static char *synopsis = "Say text to the user, using Flite";

static char *descrip = 
"  Flite(text[|intkeys]):  This will invoke the Flite TTS engine,\n"
"send a text string, get back the resulting waveform and play it to\n"
"the user, allowing any given interrupt keys to immediately terminate\n"
"and return.\n";

#if ASTERISK_VERSION < 4
STANDARD_LOCAL_USER;

LOCAL_USER_DECL;
#endif

static int flite_exec(struct ast_channel *chan, void *data)
{
	int res = 0; /* Application result */
#if ASTERISK_VERSION < 4
	struct localuser *u;
#else
	struct ast_module_user *u;
#endif

#if ASTERISK_VERSION >= 6 
        struct ast_flags config_flags = { 0 };
#endif

	char *intstr = "";	/* String containing DTMF string to interrupt on, can be AST_DIGIT_ANY */
	const char *tmp; /* Contains duplicate of <data>, for modification */
	struct ast_config *cfg; /* Asterisk config file struct */
	int useCache;
	const char *cacheDir;
	char cacheFile[2048] = "";
	char cmd[255];
	/* For cache */
	struct MD5Context md5ctx;
	unsigned char MD5Res[16];
	char MD5Hex[33] = "";
	int fdesc = -1;
	int writeCache = 0;
	char koko[4] = "";
	int i;
	char filename[27];
	char wavFilename[27];
	cst_voice *v;

	if (ast_strlen_zero(data))
	{
		ast_log(LOG_WARNING, "Flite requires an argument (text)\n");
		return -1;
	}

#if ASTERISK_VERSION >= 4
	u = ast_module_user_add(chan);
#else
	LOCAL_USER_ADD(u);
#endif

	/* Config file stuff */
#if ASTERISK_VERSION >= 6 
        cfg = ast_config_load(FLITE_CONFIG, config_flags);
#else
	cfg = ast_config_load(FLITE_CONFIG);
#endif

	if (!cfg) 
	{
		ast_log(LOG_WARNING, "Flite: No such configuration file %s\n", FLITE_CONFIG);
		useCache = 0;
		cacheDir = NULL;
	}
	else
	{
		if (!(tmp = ast_variable_retrieve(cfg, "general", "usecache")))
		{
			useCache = 0;
		}
		else
		{
			useCache = 1;
		}

		if (!(cacheDir = ast_variable_retrieve(cfg, "general", "cachedir")))
		{
			cacheDir = "/tmp";
		}
	}
	/* End of config file stuff */
	
	/* Create a duplicate of the input data */
	tmp = ast_strdupa(data);

	/* Get the allowable dtmf bargein keys */
	if ((intstr = strchr(data, '|'))) 
	{
		*intstr = '\0';
		intstr++;
		if (!strcasecmp(intstr, "any"))
			intstr = AST_DIGIT_ANY;
	}

	ast_log(LOG_DEBUG, "Text passed to Flite: %s\n", (char *) data);

	/* Prepare the cache mechansim */
	if (useCache)
	{
		/* Compute MD5 sum of string */
		MD5Init(&md5ctx);
		MD5Update(&md5ctx,(unsigned char const *)data,strlen(data));
		MD5Final(MD5Res,&md5ctx);
		MD5Hex[0] = '\0';

		/* Convert to HEX and look if there is any matching file in the cache 
    		directory */
		for (i =0; i<16; i++)
		{
			snprintf(koko, sizeof(koko), "%X", MD5Res[i]);
			strncat(MD5Hex, koko, sizeof(MD5Hex) - strlen(MD5Hex) - 1);
		}

		if (strlen(cacheDir)+strlen(MD5Hex)+5 <= 2048)
		{
			ast_log(LOG_DEBUG, "Flite: Activating cache mechanism...");
			snprintf(cacheFile, sizeof(cacheFile), "%s/%s.wav", cacheDir, MD5Hex);
			fdesc = open(cacheFile, O_RDONLY);
			cacheFile[(strlen(cacheFile)-4)] = '\0';
			if (fdesc == -1)
			{
				ast_log(LOG_DEBUG, "Flite: Cache file does not yet exist.");
				writeCache = 1;
			}
			else
			{
				close(fdesc);
				ast_log(LOG_DEBUG,"Flite: Cache file exists.");
				res = ast_streamfile(chan, cacheFile, chan->language);
		
				if (!res)
				{
					res = ast_waitstream(chan, intstr);
					ast_stopstream(chan);
				}
				else
				{
					ast_log(LOG_WARNING, "ast_streamfile failed on %s for Flite\n", chan->name);
				}

#if ASTERISK_VERSION >= 4
				ast_module_user_remove(u);
#else
				LOCAL_USER_REMOVE(u);
#endif
				return res;
			}
		}
	}
	/* End of cache mechanism */

	/* Create temp filename */
	sprintf(filename, "/tmp/flite_buf_XXXXXX");
	mkstemp(filename);
	strcpy(wavFilename, filename);
	strcat(wavFilename, ".wav");

	/* Invoke Flite */
	flite_init();
	v = register_cmu_us_kal();
	flite_text_to_speech((char *) data, v, wavFilename);

	if (chan->_state != AST_STATE_UP)
	{
		/* Answer the channel if it is down */
		ast_answer(chan);
	}

	/* Save audio file to cache, if required */
	if (writeCache)
	{
		ast_log(LOG_DEBUG, "Flite: Creating cache file...");
		sprintf(cmd, "cp -f %s %s.wav", wavFilename, cacheFile);
		system(cmd);
	}

	/* Play back the audio file */
	ast_stopstream(chan);
	res = ast_streamfile(chan, filename, chan->language);
	if (!res)
	{
		res = ast_waitstream(chan, intstr);
		ast_stopstream(chan);
	}
	else
	{
		ast_log(LOG_WARNING, "ast_streamfile failed on %s for Flite\n", chan->name);
	}

	/* Remove temp file */
	sprintf(cmd, "rm -f %s*", filename);
	system(cmd);

#if ASTERISK_VERSION >= 4
	ast_module_user_remove(u);
#else
	LOCAL_USER_REMOVE(u);
#endif
	return res;
}

#if ASTERISK_VERSION >= 4
static int unload_module(void)
{
	int res;
	res = ast_unregister_application(app);
	return res;	
}

static int load_module(void)
{
	return ast_register_application(app, flite_exec, synopsis, descrip);
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Flite TTS Interface");

#else
int unload_module(void)
{
	int res;
	res = ast_unregister_application(app);
	STANDARD_HANGUP_LOCALUSERS;
	return res;	
}

int load_module(void)
{
	return ast_register_application(app, flite_exec, synopsis, descrip);
}


char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
#endif
